/// <reference path="run_blackbox_tests.d.ts"/>

import * as chai from "chai";
import { expect } from "chai";
import * as chaiJestSnapshot from "chai-jest-snapshot";
import * as fs from "fs";
import * as glob from "glob";
import * as nock from "nock";
import * as path from "path";
import * as sinon from "sinon";
import ConnectionHelper from "./connection_helper";
import { get } from 'lodash';

chai.use(chaiJestSnapshot);

const printServiceConfig = serviceDetails => {
  const { url, port, method, pathRegex } = serviceDetails;
  console.log("### Service details");
  console.log(serviceDetails);
};

const generateTestName = (queryName, variant) => {
  return {
    ...variant,
    name: `${queryName} (${variant.name})`
  };
};

// export type GraphQLFunction = ()
export type CleanupFunction = () => void;

export type DebugOptions = {
  printResult?: boolean
}

export interface BlackboxTestsRunnerConfig {
  graphqlFn: any
  cleanupFn?: CleanupFunction
  serviceConfig: object
  servicePaths: object
  fixturesDir: string
  exclude?: string[]
  include?: string[]
  debug?: DebugOptions
}

const blackboxTestsRunner = (opts: BlackboxTestsRunnerConfig) => {
  const {
    graphqlFn,
    serviceConfig,
    servicePaths,
    fixturesDir,
    cleanupFn,
    exclude = [],
    include = [],
    debug
  } = opts;

  const files = glob.sync(`${fixturesDir}/*.graphql`);

  const connHelper = new ConnectionHelper(serviceConfig, servicePaths);

  const setUpMocks = mocks => {
    mocks.forEach(mock => {
      connHelper.mockServicePath(mock.service_path, mock.data);
    });
  };

  describe("Black box tests", () => {
    let clock = null;

    beforeEach(function() {
      cleanupFn();
      chaiJestSnapshot.configureUsingMochaContext(this);
    });

    afterEach(() => {
      nock.cleanAll();
    });

    files.forEach(file => {
      const queryName = path.basename(file, ".graphql");

      if (include.length > 0 && !include.includes(queryName)) {
        return;
      }
      if (exclude.includes(queryName)) {
        console.log(`-> Ignored ${queryName}`);
        return;
      }

      const config = JSON.parse(
        fs.readFileSync(`${fixturesDir}/${queryName}.config.json`, "utf8")
      );
      const testsInFile = config.variants
        ? config.variants.map(variant => generateTestName(queryName, variant))
        : [{ name: `${queryName}`, mocks: config.mocks, other: config.other }];
      testsInFile.map(test => {
        it(`${test.name}`, async () => {
          const query = fs.readFileSync(file, "utf8");
          setUpMocks(test.mocks);

          const fakeTimer = get(test, 'other.fakeTimer', 0);
          if (fakeTimer) {
            clock = sinon.useFakeTimers({ now: fakeTimer });
          }

          const result = await graphqlFn(query, test.variables || {});
          chaiJestSnapshot.setFilename(`${fixturesDir}/${queryName}.snap`);
          if (!nock.isDone()) {
            console.log("Pending mocks");
            console.log(nock.pendingMocks());
          }
          if (debug.printResult) {
            console.log(JSON.stringify(result, null, 2));
          }
          expect(nock.isDone()).to.equal(
            true,
            "Not all mocked endpoints were hit"
          );
          expect(result).to.matchSnapshot();
        });
      });
    });
  });
};

export default blackboxTestsRunner;
